// Distributed Boids example in RakNet
// GLUT visualiser
//
// Anthony Steed, 2009
//
// Originally based on the ReplicaManager2 example from RakNet
//

#include "StringTable.h"
#include "RakPeerInterface.h"
#include "RakNetworkFactory.h"
#include <stdio.h>
#include "Kbhit.h"
#include <string.h>
#include <ctype.h>
#include "BitStream.h"
#include "MessageIdentifiers.h"
#include "ReplicaManager2.h"
#include "NetworkIDManager.h"
#include "RakSleep.h"
#include "FormatString.h"
#include "StringCompressor.h"
#include "BoidObjects.h"

#include "GL/glut.h"
#include <float.h>
#include <math.h>

#ifndef M_PI
#define M_PI       3.14159265358979323846
#endif

#define SQRT3      1.7320508075688772935


using namespace RakNet;

NetworkIDManager networkIdManager;
bool isServer;
RakPeerInterface *rakPeer;
ReplicaManager2 replicaManager;
ReplicaManager2BoidConnectionFactory connectionFactory;

int numberBoids=10;
int frequency=10;
extern Vec3f world_min_pt;
extern Vec3f world_max_pt;

// The arrays of boids
extern DataStructures::List<Boid*> myBoids;
extern DataStructures::List<Boid*> otherBoids;

// forward declarations so we can have the interesting stuff upfront
void quitRaknet(void);
int setupGLUT(int *argc, char *argv[] );

//
// View control
//

float world_rot_y=0.0f;
float world_rot_x=30.0f;
int last_drag_x=0;
int last_drag_y=0;
float scale=1.0f;


void drawGround()
{
	int color = 0;
	float x = world_min_pt[0];
	float y = world_min_pt[1];
	float z = world_min_pt[2];

#define GROUND_TILES 10

	float dx = (world_max_pt[0] - world_min_pt[0])/GROUND_TILES;
	float dz = (world_max_pt[2] - world_min_pt[2])/GROUND_TILES;

	for (int i=0;i<GROUND_TILES;i++)
	{
		color += ((GROUND_TILES+1)%2);
		for (int j=0;j<GROUND_TILES;j++)
		{
			if ((++color)%2 == 0)
				glColor3f(0.5,0.5,0.5);
			else
				glColor3f(0.75,0.75,0.75);
			glBegin(GL_QUADS);
			glVertex3f(x,y,z);
			glVertex3f(x+dx,y,z);
			glVertex3f(x+dx,y,z+dz);
			glVertex3f(x,y,z+dz);
			glEnd();
			z += dz;
		}
		z = world_min_pt[2];
		x += dx;
	}
}


//
// The vertex/geometry specifications 
//

#define NUM_FACES 6

#define NUM_VERTS 5

float verts[NUM_VERTS*3] = {
	0.0f,  0.0f,  0.0f, // 0
	0.0f,  0.0f, -1.0f, // 1
	0.8f,  0.3f,  0.4f, // 2
	-0.8f,  0.3f,  0.4f, // 3
	0.0f, -0.1f,  0.0f, // 4
};

#define NUM_INDICES 18

int indices[NUM_INDICES] = {
	2,1,0,
	1,3,0,
	4,3,1,
	2,4,1,
	0,4,2,
	0,3,4,
};

void drawBoids(DataStructures::List<Boid*> &boids, float r, float g, float b)
{
	for (unsigned int i=0; i < boids.Size(); i++) 
	{
		glPushMatrix();
		glTranslatef(boids[i]->pos[0],boids[i]->pos[1],boids[i]->pos[2]);

		float rotateAngle = (-180.0f/M_PI)*atan2(boids[i]->vel[0], -boids[i]->vel[2]); 
		glRotatef(rotateAngle, 0.0f, 1.0f, 0.0f);

		glColor3f(r,g,b);

		glBegin(GL_TRIANGLES);
		for (int i=0;i<NUM_INDICES;i++)
		{
			glVertex3f(verts[indices[i]*3],verts[indices[i]*3+1],verts[indices[i]*3+2]); 
		}
		glEnd();
		glPopMatrix();	
	}
}

// Draw our scene
void drawScene()
{
	glClearColor(1.0, 1.0, 1.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


	glPushMatrix();
	glRotatef(world_rot_x, 1.0f,0.0f,0.0f);
	glRotatef(world_rot_y, 0.0f,1.0f,0.0f);
	glScalef(scale,scale,scale);

	drawGround();

	drawBoids(myBoids, 0,1,0);
	drawBoids(otherBoids, 1,0,0);

	glPopMatrix();
}


// update loop
void update( void )
{
	Packet *packet;

	updateBoids(1.0f/frequency);

	for (packet = rakPeer->Receive(); packet; rakPeer->DeallocatePacket(packet), packet = rakPeer->Receive())
	{
		switch (packet->data[0])
		{
		case ID_CONNECTION_ATTEMPT_FAILED:
			printf("ID_CONNECTION_ATTEMPT_FAILED\n");
			quitRaknet();
			break;
		case ID_NO_FREE_INCOMING_CONNECTIONS:
			printf("ID_NO_FREE_INCOMING_CONNECTIONS\n");
			quitRaknet();
			break;
		case ID_CONNECTION_REQUEST_ACCEPTED:
			printf("ID_CONNECTION_REQUEST_ACCEPTED\n");
			replicaManager.AddNewConnection(packet->systemAddress);
			break;
		case ID_NEW_INCOMING_CONNECTION:
			printf("ID_NEW_INCOMING_CONNECTION from %s\n", packet->systemAddress.ToString());
			replicaManager.AddNewConnection(packet->systemAddress);
			break;
		case ID_DISCONNECTION_NOTIFICATION:
			printf("ID_DISCONNECTION_NOTIFICATION\n");
			Boid::DeleteBoidsByAddress(packet->systemAddress);
			break;
		case ID_CONNECTION_LOST:
			printf("ID_CONNECTION_LOST\n");
			Boid::DeleteBoidsByAddress(packet->systemAddress);
			break;
		}
	}

	drawScene(); // Draws the world including all boids

	glutSwapBuffers(); // Indicate to GLUT to refresh screen

	RakSleep((int)(1000.0/frequency));
}

int main(int argc, char *argv[])
{
	SocketDescriptor sd;
	char ip[128];
	static const int SERVER_PORT=12345;

	int winid = setupGLUT(&argc, argv); // Start GLUT

	rakPeer = RakNetworkFactory::GetRakPeerInterface();
	isServer=false;
	sd.port=0;

	rakPeer->SetNetworkIDManager(&networkIdManager);
	networkIdManager.SetIsNetworkIDAuthority(isServer);
	rakPeer->Startup(isServer ? 32 : 1,100,&sd,1);
	rakPeer->AttachPlugin(&replicaManager);
	replicaManager.SetAutoAddNewConnections(false);
	replicaManager.SetConnectionFactory(&connectionFactory);

	strcpy(ip, "127.0.0.1");
	rakPeer->Connect(ip,SERVER_PORT,0,0,0);

	RakNet::StringTable::Instance()->AddString("Boid", false);

	printf("\nCommands:\n(Q)uit\n(P)rint out all objects\n(S)pawn / despawn Boid\n\n");

	glutMainLoop();	// GLUT main loop

}

void quitRaknet()
{
	rakPeer->Shutdown(100,0);
	RakNetworkFactory::DestroyRakPeerInterface(rakPeer);
}

//
// GLUT callback functions
//

// react to window size changes
void reshape(int w, int h)
{
	glMatrixMode(GL_PROJECTION);
	glViewport(0, 0, w, h);
	glMatrixMode(GL_MODELVIEW);
	glutPostRedisplay();
}

// We change camera angle on mouse drags
void mouse(int button, int state, int x, int y)
{
	last_drag_x = x;
	last_drag_y = y;
	glutPostRedisplay();
}

// We change camera angle on mouse drags
void motion(int x, int y)
{
	world_rot_y += (x-last_drag_x)/5.0f;
	world_rot_x += (y-last_drag_y)/5.0f;
	if (world_rot_x > 90.0f) world_rot_x = 90.0f;
	if (world_rot_x < 0.0f) world_rot_x = 0.0f;
	last_drag_x = x;
	last_drag_y = y;

	glutPostRedisplay();
}

// React to key presses
void keyboard(unsigned char k, int x, int y)
{
	switch(k)
	{
	case '+':
		scale *=1.5;
		break;
	case '-':
		scale *=0.75;
		break;
	case 'q':
	case 27:    
		printf("Quitting.\n");
		quitRaknet();
		break;
	case 'p':
		printAll();
		break;
	case 's':
		if (!isServer)
		{
			if (myBoids.Size()!=0)
			{
				printf("Destroy Boids.\n");
				while (myBoids.Size()>0)
				{
					myBoids[0]->BroadcastDestruction();
					delete myBoids[0];
				}
			}
			else
			{
				printf("Create Boids.\n");
				for (int i=0;i<numberBoids;i++)
				{
					Boid* newBoid = new Boid();
					myBoids.Insert(newBoid, __FILE__, __LINE__);
					newBoid->SetReplicaManager(&replicaManager);
					newBoid->AddAutoSerializeTimer(100);
					newBoid->BroadcastConstruction();
				}
			}
		}
		break;
	}
}

// Setup GLUT, create a Window and initialize callbacks
int setupGLUT(int *argc, char *argv[])
{
	glutInit(argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);

	int winid = glutCreateWindow("OpenGL Boids");

	glutReshapeFunc(reshape);
	glutDisplayFunc(update);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutKeyboardFunc(keyboard);
	glutIdleFunc(update);

	// The camera is conservative so the boids are all always on screen
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(SQRT3*world_min_pt[0], SQRT3*world_max_pt[0],
		SQRT3*world_min_pt[1], SQRT3*world_max_pt[1],
		0.1, 100*world_max_pt[2]);

	return winid;
}
